"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Node_1 = require("./Node");
const Subject_1 = require("./Subject");
const Resource_1 = require("./Resource");
const util_1 = require("../util");
/**
  Class for building Resource/Subject class hierarchy based on schema
 */
class Graph {
    constructor(schema) {
        this.schema = schema;
        this.subjectChildMap = new Map();
        this.resourceChildMap = new Map();
        this.resources = Graph.buildResourceHierachy(schema.resources);
        this.subjects = Graph.buildSubjectHierachy(schema.subjects);
        // invert resource graph
        this.resources.forEach((ChildResource, childName) => {
            const parentResources = ChildResource.getHierarchyClassNames();
            parentResources.forEach(parentName => {
                if (!this.resourceChildMap.has(parentName))
                    this.resourceChildMap.set(parentName, new Map());
                if (parentName !== childName)
                    this.resourceChildMap.get(parentName).set(childName, ChildResource);
            });
        });
        // invert subject graph
        this.subjects.forEach((ChildSubject, childName) => {
            const parentSubjects = ChildSubject.getHierarchyClassNames();
            parentSubjects.forEach(parentName => {
                if (!this.subjectChildMap.has(parentName))
                    this.subjectChildMap.set(parentName, new Map());
                if (parentName !== childName)
                    this.subjectChildMap.get(parentName).set(childName, ChildSubject);
            });
        });
    }
    /**
     *  Build map of nodeName -> ResourceClass for gracl hierarchy
        NOTE: unfortunately we need to duplicate this function for subjects
              due to limitations of the way type parameters are handled
          See: https://github.com/Microsoft/TypeScript/issues/4890
     */
    static buildResourceHierachy(schemaNodes) {
        const nodeList = util_1.topologicalSort(schemaNodes), classGraph = new Map();
        const createClass = (node) => {
            const getParentsMethod = node.getParents || Node_1.Node.prototype.getParents;
            const getPermissionMethod = node.getPermission || Resource_1.Resource.prototype.getPermission;
            if (node.parent) {
                const ParentClass = classGraph.get(node.parent);
                classGraph.set(node.name, (_a = class extends ParentClass {
                        getParents() {
                            return getParentsMethod.call(this);
                        }
                        getPermission(subject) {
                            return getPermissionMethod.call(this, subject);
                        }
                    },
                    _a.id = node.id || 'id',
                    _a.parentId = node.parentId,
                    _a.displayName = node.name,
                    _a.repository = node.repository,
                    _a.permissionPropertyKey = node.permissionProperty ||
                        'permissions',
                    _a));
            }
            else {
                classGraph.set(node.name, (_b = class extends Resource_1.Resource {
                        getPermission(subject) {
                            return getPermissionMethod.call(this, subject);
                        }
                    },
                    _b.id = node.id || 'id',
                    _b.displayName = node.name,
                    _b.repository = node.repository,
                    _b.permissionPropertyKey = node.permissionProperty ||
                        'permissions',
                    _b));
            }
            var _a, _b;
        };
        nodeList.forEach(createClass);
        return classGraph;
    }
    /**
     *  Build map of nodeName -> SubjectClass for gracl hierarchy
        NOTE: this is a duplicate of buildResourceHierachy that
              is necessary for reasons explained above due to limitations
              in the type system.
     */
    static buildSubjectHierachy(schemaNodes) {
        const nodeList = util_1.topologicalSort(schemaNodes), classGraph = new Map();
        const createClass = (node) => {
            const getParentsMethod = node.getParents || Node_1.Node.prototype.getParents;
            if (node.parent) {
                const ParentClass = classGraph.get(node.parent);
                classGraph.set(node.name, (_a = class extends ParentClass {
                        getParents() {
                            return getParentsMethod.call(this);
                        }
                    },
                    _a.id = node.id || 'id',
                    _a.parentId = node.parentId,
                    _a.displayName = node.name,
                    _a.repository = node.repository,
                    _a.permissionPropertyKey = node.permissionProperty ||
                        'permissions',
                    _a));
            }
            else {
                classGraph.set(node.name, (_b = class extends Subject_1.Subject {
                    },
                    _b.id = node.id || 'id',
                    _b.displayName = node.name,
                    _b.repository = node.repository,
                    _b.permissionPropertyKey = node.permissionProperty ||
                        'permissions',
                    _b));
            }
            var _a, _b;
        };
        nodeList.forEach(createClass);
        return classGraph;
    }
    static resolveNodeName(node) {
        return typeof node === 'string' ? node : node.displayName;
    }
    /**
     *  Extract a node class from the graph if it exists
     */
    getClass(node, type) {
        const name = Graph.resolveNodeName(node);
        let map;
        switch (type) {
            case 'subject':
                map = this.subjects;
                break;
            case 'resource':
                map = this.resources;
                break;
            default:
                throw new Error(`Invalid class type ${type}, must be subject or resource!`);
        }
        if (map.has(name))
            return map.get(name);
        throw new Error(`No ${type} class found for ${name}!`);
    }
    /**
     *  Extract a resource class from the graph if it exists
     */
    getResource(node) {
        const name = Graph.resolveNodeName(node);
        return this.getClass(name, 'resource');
    }
    /**
     *  Extract a subject class from the graph if it exists
     */
    getSubject(node) {
        const name = Graph.resolveNodeName(node);
        return this.getClass(name, 'subject');
    }
    getChildResources(node) {
        const name = Graph.resolveNodeName(node);
        if (!this.resources.has(name))
            throw new Error(`No resource class found for ${name}!`);
        return Array.from(this.resourceChildMap.get(name).values());
    }
    getChildSubjects(node) {
        const name = Graph.resolveNodeName(node);
        if (!this.subjects.has(name))
            throw new Error(`No subject class found for ${name}!`);
        return Array.from(this.subjectChildMap.get(name).values());
    }
    getParentResources(node) {
        const name = Graph.resolveNodeName(node);
        const ResourceClass = this.getResource(name);
        return ResourceClass.getHierarchyClassNames()
            .filter(n => n !== name)
            .map(n => this.getResource(n));
    }
    getParentSubjects(node) {
        const name = Graph.resolveNodeName(node);
        const SubjectClass = this.getSubject(name);
        return SubjectClass.getHierarchyClassNames()
            .filter(n => n !== name)
            .map(n => this.getSubject(n));
    }
}
exports.Graph = Graph;
